import copy

from .commons import deep_get
from .mappers import Constant, Deleted
from .structures import Structure
from .fields import FunctionCall, PositiveInt


VERSION_MAPPING = "_versions_mapping"


class Versioned(Structure):
    """
    Marks a structure as can be deserialized from multiple versions.
    The version is expected to start with 1 and increase by 1 in every update.
    It is expected to have a class attribute of "_versions_mapping", with an ordered list
    of the mappings. The first mapping maps version 1 to 2, the second 2 to 3, etc.

    Arguments:

    _versions_mapping: optional
        An array of mappers that outlines how to convert older versions to the latest version.

    """

    version: PositiveInt

    def __init__(self, *args, **kwargs):
        versions_mapping = getattr(self, VERSION_MAPPING, [])
        default_version = len(versions_mapping) + 1
        kwargs["version"] = default_version
        super().__init__(*args, **kwargs)


def _convert(mapped_dict: dict, mapping):
    out_dict = copy.deepcopy(mapped_dict)
    for k, v in mapping.items():
        if isinstance(v, Constant):
            out_dict[k] = v()
        elif k.endswith("._mapper"):
            field_name = k[: -len("._mapper")]
            content = mapped_dict.get(field_name, None)
            if content is not None:
                if isinstance(content, list):
                    out_dict[field_name] = [_convert(x, v) for x in content]
                else:
                    out_dict[field_name] = _convert(content, v)

        elif isinstance(v, FunctionCall):
            args = [out_dict.get(x) for x in v.args] if v.args else [out_dict.get(k)]
            out_dict[k] = v.func(*args)

    for k, v in mapping.items():
        if isinstance(v, str):
            out_dict[k] = deep_get(out_dict, v)

    for k, v in mapping.items():
        if v == Deleted and k in out_dict:
            del out_dict[k]
    return out_dict


def convert_dict(the_dict: dict, versions_mapping):
    start_version = the_dict.get("version", 1)
    mapped_dict = copy.deepcopy(the_dict)
    for mapping in versions_mapping[(start_version - 1) :]:
        mapped_dict = _convert(mapped_dict, mapping)
        mapped_dict["version"] = mapped_dict.get("version", 0) + 1
    return mapped_dict



import collections.abc
import enum
import importlib.util
import inspect
import logging
import os
import sys
import typing
from os.path import relpath
from pathlib import Path
from .fields import AnyOf, FunctionCall, Map
from .enum import Enum
from .serialization_wrappers import Deserializer, Serializer
from .structures import (
    ImmutableStructure,
    NoneField,
    Structure,
    Field,
    get_typing_lib_info,
)
from .utility import type_is_generic
import builtins

builtins_types = {getattr(builtins, k) for k in dir(builtins) if isinstance(getattr(builtins, k), type)}

INDENT = " " * 4

AUTOGEN_NOTE = [
    "",
    "#### This stub was autogenerated by Typedpy",
    "###########################################",
    "",
]


def _get_package(v, attrs):
    pkg_name = attrs.get("__package__") or "%^$%^$%^#"
    if v.startswith(pkg_name):
        return v[len(pkg_name):]
    return v


def _as_something(k, attrs):
    return f" as {k}" if attrs.get("__file__", "").endswith("__init__.py") else ""


def _get_anyof_typing(field, locals_attrs, additional_classes):
    union_fields = getattr(field, "_fields", [])
    if len(union_fields) == 2 and isinstance(union_fields[1], NoneField):
        info = _get_type_info(union_fields[0], locals_attrs, additional_classes)
        return f"Optional[{info}] = None"

    fields = ",".join(
        [_get_type_info(f, locals_attrs, additional_classes) for f in union_fields]
    )
    return f"Union[{fields}]"


def _get_type_info(field, locals_attrs, additional_classes):
    if isinstance(field, AnyOf):
        return _get_anyof_typing(field, locals_attrs, additional_classes)

    if isinstance(field, Map):
        if not field.items:
            return "dict"
        sub_types = ", ".join(
            [_get_type_info(f, locals_attrs, additional_classes) for f in field.items]
        )
        return f"dict[{sub_types}]"

    if isinstance(field, Enum):
        if field._is_enum:
            additional_classes.add(field._enum_class)
            return field._enum_class.__name__

    the_type = getattr(field, "get_type", field) if isinstance(field, Field) else field
    if the_type in builtins_types:
        return the_type.__name__

    if the_type is typing.Any:
        return "Any"
    if (
        not the_type.__module__.startswith("typedpy")
        and the_type.__module__ != "builtins"
        and the_type not in locals_attrs
        and not type_is_generic(the_type)
    ):
        additional_classes.add(field)
    if type_is_generic(the_type):
        origin = getattr(the_type, "__origin__", None)
        args = getattr(the_type, "__args__", None)
        if origin in {list, set, tuple}:
            if len(args) != 1:
                return str(origin)
            return f"{origin.__name__}[{_get_type_info(args[0], locals_attrs, additional_classes)}]"

        if origin is collections.abc.Callable:
            mapped_args = (
                [_get_type_info(a, locals_attrs, additional_classes) for a in args]
                if args
                else []
            )
            args_st = (
                ""
                if not mapped_args
                else f"[{mapped_args[0]}]"
                if len(mapped_args) == 1
                else f"[[{','.join(mapped_args[:-1])}], {mapped_args[-1]}]"
            )
            # args_st = "" if not mapped_args else f"[{','.join(mapped_args)}]"
            additional_classes.add(typing.Callable)
            return f"Callable{args_st}"

        if origin is collections.abc.Iterable:
            additional_classes.add(typing.Iterable)
            mapped_args = (
                [_get_type_info(a, locals_attrs, additional_classes) for a in args]
                if args
                else []
            )
            args_st = (
                ""
                if not mapped_args
                else f"[{mapped_args[0]}]"
            )
            return f"Iterable{args_st}"

        return _get_type_info(
            get_typing_lib_info(field), locals_attrs, additional_classes
        )

    return f"{the_type.__name__}"


def _get_all_type_info(cls, locals_attrs, additional_classes) -> dict:
    type_by_name = {}
    required = getattr(cls, "_required", None)
    for field_name, field in cls.get_all_fields_by_name().items():
        try:
            type_info_str: str = _get_type_info(field, locals_attrs, additional_classes)
            if (
                field_name not in required
                and required is not None
                and not type_info_str.startswith("Optional[")
            ):
                type_info_str = f"Optional[{type_info_str}] = None"
            type_by_name[field_name] = type_info_str
        except Exception as e:
            logging.exception(e)

    return type_by_name


def _get_struct_classes(attrs, only_calling_module=True):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and issubclass(v, Structure)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
            and v not in {Deserializer, Serializer, ImmutableStructure, FunctionCall}
        )
    }


def _get_imported_classes(attrs):
    return [
        f"from {_get_package(v.__module__, attrs)} import {k}{_as_something(k, attrs)}"
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and attrs["__name__"] != v.__module__
            and not v.__module__.startswith("typing")
            and not v.__module__.startswith("typedpy")
        )
    ]


def _get_ordered_args(unordered_args: dict):
    optional_args = {k: v for k, v in unordered_args.items() if v.endswith("= None")}
    mandatory_args = {k: v for k, v in unordered_args.items() if k not in optional_args}
    return {**mandatory_args, **optional_args}


def _get_mapped_extra_imports(additional_imports) -> dict:
    mapped = {}
    for c in additional_imports:
        try:
            name = _get_type_info(c, {}, set())
            if inspect.isclass(c) and issubclass(c, Structure):
                module_name = c.__module__
            else:
                module_name = (
                    c.get_type.__module__
                    if isinstance(c, Field)
                    or (inspect.isclass(c) and issubclass(c, Field))
                    else c.__module__
                )
            mapped[name] = module_name
        except Exception as e:
            logging.exception(e)
    return mapped


def _get_methods_info(cls, locals_attrs, additional_classes) -> list:
    method_by_name = []
    all_fields = cls.get_all_fields_by_name() if issubclass(cls, Structure) else {}
    ignored_methods = (
        dir(Structure)
        if issubclass(cls, Structure)
        else dir(Field)
        if issubclass(cls, Structure)
        else dir(enum)
        if issubclass(cls, enum.Enum)
        else {}
    )
    mros = list(reversed(inspect.getmro(cls)))[1:]
    members = {}
    for c in mros:
        members.update(dict(c.__dict__))
    method_list = [
        attribute
        for attribute in members
        if (
            callable(getattr(cls, attribute, None))
            or isinstance(getattr(cls, attribute, None), property)
        )
        and not attribute.startswith("_")
        and attribute not in all_fields
        and attribute not in ignored_methods
    ]
    if (
        not issubclass(cls, Structure)
        and not issubclass(cls, enum.Enum)
        and "__init__" in members
    ):
        method_list = ["__init__"] + method_list

    for name in method_list:
        method_cls = members[name].__class__
        is_property = False
        func = getattr(cls, name)
        if isinstance(func, property):
            is_property = True
            func = func.__get__
        sig = inspect.signature(func)
        return_annotations = (
            ""
            if sig.return_annotation == inspect._empty
            else f" -> {_get_type_info(sig.return_annotation, locals_attrs, additional_classes)}"
        )
        params_by_name = {}
        if method_cls is classmethod:
            params_by_name["cls"] = ""
        for p, v in sig.parameters.items():
            optional_globe = (
                "**"
                if v.kind == inspect.Parameter.VAR_KEYWORD
                else "*"
                if v.kind == inspect.Parameter.VAR_POSITIONAL
                else ""
            )
            default = "" if v.default == inspect._empty else f" = {v.default}"
            type_annotation = (
                ""
                if v.annotation == inspect._empty
                else f": {_get_type_info(v.annotation, locals_attrs, additional_classes)}"
            )
            p_name = f"{optional_globe}{p}"
            params_by_name[p_name] = f"{type_annotation}{default}"
        if is_property:
            params_by_name.pop("owner")
            params_by_name.pop("instance")
            params_by_name["self"] = ""
        params_as_str = ", ".join([f"{k}{v}" for k, v in params_by_name.items()])
        method_by_name.append("")
        if method_cls is staticmethod:
            method_by_name.append("@staticmethod")
        elif method_cls is classmethod:
            method_by_name.append("@classmethod")
        elif is_property:
            method_by_name.append("@property")
        method_by_name.append(f"def {name}({params_as_str}){return_annotations}: ...")

    return method_by_name


def _get_init(cls, ordered_args: dict) -> str:
    init_params = f",\n{INDENT * 2}".join(
        [f"{INDENT * 2}self"] + [f"{k}: {v}" for k, v in ordered_args.items()]
    )
    kw_opt = (
        f",\n{INDENT * 2}**kw" if getattr(cls, "_additionalProperties", True) else ""
    )
    return f"    def __init__(\n{init_params}{kw_opt}\n{INDENT}): ..."


def get_stubs_of_structures(
    struct_classe_by_name: dict, local_attrs, additional_classes
) -> list:
    out_src = []
    for cls_name, cls in struct_classe_by_name.items():
        fields_info = _get_all_type_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )

        if not fields_info and not method_info:
            continue

        ordered_args = _get_ordered_args(fields_info)
        out_src.append(f"class {cls_name}(Structure):")
        out_src.append(_get_init(cls, ordered_args))
        out_src.append("")

        for field_name, type_name in ordered_args.items():
            out_src.append(f"    {field_name}: {type_name}")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def get_stubs_of_enums(
    enum_classes_by_name: dict, local_attrs, additional_classes
) -> list:
    out_src = []
    for cls_name, cls in enum_classes_by_name.items():

        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )

        out_src.append(f"class {cls_name}(enum.Enum):")
        for v in cls:
            out_src.append(f"{INDENT}{v.name} = enum.auto()")
        out_src.append("")

        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def add_imports(local_attrs: dict, additional_classes, existing_imports: set) -> list:
    base_import_statements = [
        "from typing import Union, Optional, Any",
        "from typedpy import Structure",
        "import enum",
        "",
    ]
    extra_imports_by_name = _get_mapped_extra_imports(additional_classes)
    extra_imports = {
        f"from {_get_package(v, local_attrs)} import {k}{_as_something(k, local_attrs)}"
        for k, v in extra_imports_by_name.items()
        if (
            k not in local_attrs or local_attrs[k].__module__ != local_attrs["__name__"]
        )
    } - existing_imports
    return base_import_statements + sorted(extra_imports)


def _get_enum_classes(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and issubclass(v, enum.Enum)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_other_classes(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and not issubclass(v, enum.Enum)
            and not issubclass(v, Structure)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_functions(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isfunction(v)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_type_annotation(prefix: str, annotation, local_attrs, additional_classes):
    try:
        return (
            ""
            if annotation == inspect._empty
            else f"{prefix}{_get_type_info(annotation, local_attrs, additional_classes)}"
        )
    except Exception as e:
        logging.exception(e)
        return ""


def get_stubs_of_functions(func_by_name, local_attrs, additional_classes) -> list:
    out_src = []
    for name, func in func_by_name.items():
        sig = inspect.signature(func)
        return_annotations = _get_type_annotation(
            " -> ", sig.return_annotation, local_attrs, additional_classes
        )
        params_by_name = {}
        for p, v in sig.parameters.items():
            default = "" if v.default == inspect._empty else f" = {v.default}"
            type_annotation = _get_type_annotation(
                ": ", v.annotation, local_attrs, additional_classes
            )
            params_by_name[p] = f"{type_annotation}{default}"
        params_as_str = ", ".join([f"{k}{v}" for k, v in params_by_name.items()])

        out_src.append(f"def {name}({params_as_str}){return_annotations}: ...")
        out_src.append("\n")
    return out_src


def get_stubs_of_other_classes(other_classes, local_attrs, additional_classes):
    out_src = []
    for cls_name, cls in other_classes.items():
        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )

        if not method_info:
            continue

        out_src.append(f"class {cls_name}:")
        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def create_pyi(calling_source_file, attrs: dict, only_current_module: bool = True):
    full_path: Path = Path(calling_source_file)
    pyi_path = (full_path.parent / f"{full_path.stem}.pyi").resolve()
    out_src = []

    additional_imports = []
    if only_current_module:
        additional_imports = _get_imported_classes(attrs)
        if additional_imports:
            additional_imports.append("")
        out_src += additional_imports

    enum_classes = _get_enum_classes(attrs, only_calling_module=only_current_module)
    struct_classes = _get_struct_classes(attrs, only_calling_module=only_current_module)
    other_classes = _get_other_classes(attrs, only_calling_module=only_current_module)
    functions = _get_functions(attrs, only_calling_module=only_current_module)

    additional_classes = set()
    out_src += get_stubs_of_enums(
        enum_classes, local_attrs=attrs, additional_classes=additional_classes
    )
    out_src += get_stubs_of_other_classes(
        other_classes, local_attrs=attrs, additional_classes=additional_classes
    )
    out_src += get_stubs_of_structures(
        struct_classes, local_attrs=attrs, additional_classes=additional_classes
    )

    out_src += get_stubs_of_functions(
        functions, local_attrs=attrs, additional_classes=additional_classes
    )

    out_src = (
        add_imports(
            local_attrs=attrs,
            additional_classes=additional_classes,
            existing_imports=set(additional_imports),
        )
        + out_src
    )

    out_src = AUTOGEN_NOTE + out_src
    out_s = "\n".join(out_src)
    with open(pyi_path, "w", encoding="UTF-8") as f:
        f.write(out_s)


def create_stub_for_file(abs_module_path: str, src_root: str, stubs_root: str = None):
    ext = os.path.splitext(abs_module_path)[-1].lower()
    if ext != ".py":
        return
    stem = Path(abs_module_path).stem
    dir_name = str(Path(abs_module_path).parent)
    relative_dir = relpath(dir_name, src_root)
    package_name = ".".join(Path(relative_dir).parts)
    module_name = stem if stem!="__init__" else package_name
    sys.path.append(str(Path(dir_name).parent))
    spec = importlib.util.spec_from_file_location(module_name, abs_module_path)
    the_module = importlib.util.module_from_spec(spec)
    if not the_module.__package__:
        the_module.__package__ = package_name
    spec.loader.exec_module(the_module)

    pyi_dir = (
        Path(stubs_root) / Path(relative_dir)
        if stubs_root
        else Path(abs_module_path).parent
    )
    pyi_dir.mkdir(parents=True, exist_ok=True)
    (pyi_dir / Path("__init__.pyi")).touch(exist_ok=True)

    pyi_path = (pyi_dir / f"{stem}.pyi").resolve()
    if not getattr(the_module, "__package__", None):
        the_module.__package__ = ".".join(Path(relative_dir).parts)
    create_pyi(str(pyi_path), the_module.__dict__)



import enum

from .commons import first_in, wrap_val
from .fields import SerializableField, String
from .structures import Field, _FieldMeta


class _EnumMeta(_FieldMeta):
    def __getitem__(cls, values):
        if isinstance(values, (type,)) and issubclass(values, (enum.Enum,)):
            return cls(values=values)  # pylint: disable=E1120, E1123
        return cls(values=list(values))  # pylint: disable=E1120, E1123


def _all_values_from_single_enum(values):
    clazz = first_in(values).__class__
    if not (isinstance(clazz, (type,)) and issubclass(clazz, enum.Enum)):
        return False
    return all([v.__class__ is clazz for v in values])


class Enum(SerializableField, metaclass=_EnumMeta):
    """
    Enum field. value can be one of predefined values.

    Arguments:
         values(`list` or `set` or `tuple`, alternatively an enum Type):
             allowed values. Can be of any type.
             Alternatively, can be an enum.Enum type. See example below.
             When defined with an enum.Enum, serialization converts to strings,
             while deserialization expects strings (unless using serialization_by_value).
             In this case, strings are converted to the original enum values.

             Another option is assign a list of specific values from an enum.Enum class.
             In this case, it will work like asigning an Enum class, but allowing only specific values
             of that enum (see example below).

         serialization_by_value(bool): optional
             When set to True and the values is an enum.Enum class, then instead of serializing
             by the name of the enum, serialize by its value, and similarly, when deserializing,
             expect the enum value instead of the name. Default is False.
             This is especially useful when you are interfacing with another system and the values
              are strings that you don't control, like the example below.



    Examples:

    .. code-block:: python

       class Values(enum.Enum):
            ABC = 1
            DEF = 2
            GHI = 3

       class Example(Structure):
          arr = Array[Enum[Values]]
          e = Enum['abc', 'x', 'def', 3]

       example = Example(arr=[Values.ABC, 'DEF'],e=3)
       assert example.arr = [Values.ABC, Values.DEF]

       # deserialization example:
       deserialized = Deserializer(target_class=Example).deserialize({'arr': ['GHI', 'DEF', 'ABC'], 'e': 3})
       assert deserialized.arr == [Values.GHI, Values.DEF, Values.ABC]


    An example of serialization_by_value:

     .. code-block:: python

        class StreamCommand(enum.Enum):
            open = "open new stream"
            close = "close current stream"
            delete = "delete steam"

        class Action(Structure):
            command: Enum(values=StreamCommand, serialization_by_value=True)
            ....


        assert Deserializer(Action).deserialize({"command": "delete stream"}).command is StreamCommand.delete

    An example of allowing only specific values of an Enum class, referencging StreamCommand in the previous example:

    .. code-block:: python

       class Action(Structure):
            command: Enum(values=[StreamCommand.open, StreamCommand.close], serialization_by_value=True)
            ....

      # command works like in the previous example, but allows open/close values from StreamCommand, thus
      # the following line results in an exception:
      Deserializer(Action).deserialize({"command": "delete stream"})

    """

    def __init__(self, *args, values, serialization_by_value: bool = False, **kwargs):
        if not values:
            raise ValueError("Enum requires values parameters")
        self._is_enum = (
            isinstance(values, (type,))
            and issubclass(values, enum.Enum)
            or _all_values_from_single_enum(values)
        )
        self.serialization_by_value = serialization_by_value

        if self._is_enum:
            self._enum_class = (
                values if isinstance(values, (type,)) else first_in(values).__class__
            )
            if serialization_by_value:
                self._enum_by_value = {e.value: e for e in self._enum_class}
            self._valid_enum_values = (
                [v for v in self._enum_class] if isinstance(values, (type,)) else values
            )
            self.values = list(values)
        else:
            self.values = values
        super().__init__(*args, **kwargs)

    def _validate(self, value):
        if self._is_enum:
            enum_names = {v.name for v in self._valid_enum_values}
            if value not in enum_names and value not in self._valid_enum_values:
                enum_values = [r.name for r in self._valid_enum_values]
                if len(enum_values) < 11:
                    raise ValueError(
                        f"{self._name}: Got {value}; Expected one of: {', '.join(enum_values)}"
                    )
                raise ValueError(
                    f"{self._name}: Got {value}; Expected a value of {self._enum_class}"
                )

        elif value not in self.values:
            raise ValueError(
                f"{self._name}: Got {value}; Expected one of {', '.join([str(v) for v in self.values])}"
            )

    def serialize(self, value):
        if self._is_enum:
            if self.serialization_by_value:
                if not isinstance(value.value, (bool, str, int, float)):
                    raise TypeError(
                        f"{self._name}: Cannot serialize value: {value.value}"
                    )
            return value.value if self.serialization_by_value else value.name
        return value

    def deserialize(self, value):  # pylint: disable=no-self-use
        if self._is_enum:
            if self.serialization_by_value:
                if value not in self._enum_by_value:
                    raise ValueError(f"Invalid value: {wrap_val(value)}")
                return self._enum_by_value[value]
            if isinstance(value, (str,)):
                valid_names = {v.name for v in self._valid_enum_values}
                if value not in valid_names:
                    raise ValueError(f"Invalid value: {wrap_val(value)}")
                return self._enum_class[value]

        self._validate(value)
        return value

    def __set__(self, instance, value):
        self._validate(value)
        if self._is_enum:
            if isinstance(value, (str,)):
                value = self._enum_class[value]
        super().__set__(instance, value)


class EnumString(Enum, String):
    """
    Combination of :class:`Enum` and :class:`String`. This is useful if you want to further
    limit your allowable enum values, using :class:`String` attributes, such as pattern, maxLength.

    Example:

    .. code-block:: python

        predefined_list = ['abc', 'x', 'def', 'yy']

        EnumString(values=predefined_list, minLength=3)

    """

    pass



import collections
import enum
import json
import uuid
from typing import Dict

from .commons import deep_get, raise_errs_if_needed
from .versioned_mapping import VERSION_MAPPING, Versioned, convert_dict
from .mappers import (
    Constant,
    DoNotSerialize,
    aggregate_deserialization_mappers,
    aggregate_serialization_mappers,
    mappers,
)
from .structures import (
    TypedField,
    Structure,
    _get_all_fields_by_name,
    ADDITIONAL_PROPERTIES,
    REQUIRED_FIELDS,
    IGNORE_NONE_VALUES,
    NoneField,
)
from .fields import (
    Field,
    FunctionCall,
    Number,
    String,
    StructureReference,
    Array,
    Map,
    ClassReference,
    MultiFieldWrapper,
    Boolean,
    Tuple,
    Set,
    Anything,
    AnyOf,
    AllOf,
    OneOf,
    NotField,
    SerializableField,
    SizedCollection,
    wrap_val,
    _DictStruct,
    _ListStruct,
    Deque,
    Generator,
)

# pylint: disable=too-many-locals, too-many-arguments, too-many-branches


def deserialize_list_like(
    field,
    content_type,
    value,
    name,
    *,
    keep_undefined=True,
    mapper=None,
    camel_case_convert=False,
):
    if not isinstance(value, (list, tuple, set)):
        raise ValueError(f"{name}: Got {value}; Expected a list, set, or tuple")

    values = []
    items = field.items
    if isinstance(items, Field):
        ignore_none = getattr(items, IGNORE_NONE_VALUES, False)
        for i, v in enumerate(value):
            item_name = f"{name}_{i}"
            try:
                list_item = deserialize_single_field(
                    items,
                    v,
                    item_name,
                    keep_undefined=keep_undefined,
                    mapper=mapper,
                    camel_case_convert=camel_case_convert,
                    ignore_none=ignore_none,
                )
            except (ValueError, TypeError) as e:
                prefix = "" if str(e).startswith(item_name) else f"{item_name}: "
                raise ValueError(f"{prefix}{str(e)}") from e
            values.append(list_item)
    elif isinstance(items, (list, tuple)):
        for i, item in enumerate(items):
            try:
                ignore_none = getattr(item, IGNORE_NONE_VALUES, False)
                res = deserialize_single_field(
                    item,
                    value[i],
                    name,
                    keep_undefined=keep_undefined,
                    mapper=mapper,
                    camel_case_convert=camel_case_convert,
                    ignore_none=ignore_none,
                )
            except (ValueError, TypeError) as e:
                raise ValueError(f"{name}_{i}: {str(e)}") from e
            values.append(res)
        values += value[len(items) :]
    else:
        values = value
    return content_type(values)


def deserialize_array(
    array_field, value, name, *, keep_undefined=True, mapper, camel_case_convert=False
):
    return deserialize_list_like(
        array_field,
        list,
        value,
        name,
        keep_undefined=keep_undefined,
        mapper=mapper,
        camel_case_convert=camel_case_convert,
    )


def deserialize_deque(
    array_field, value, name, *, keep_undefined=True, mapper, camel_case_convert=False
):
    return deserialize_list_like(
        array_field,
        collections.deque,
        value,
        name,
        keep_undefined=keep_undefined,
        mapper=mapper,
        camel_case_convert=camel_case_convert,
    )


def deserialize_tuple(
    tuple_field, value, name, *, keep_undefined=True, mapper, camel_case_convert=False
):
    return deserialize_list_like(
        tuple_field,
        tuple,
        value,
        name,
        keep_undefined=keep_undefined,
        mapper=mapper,
        camel_case_convert=camel_case_convert,
    )


def deserialize_set(
    set_field, value, name, *, keep_undefined=True, mapper, camel_case_convert=False
):
    return deserialize_list_like(
        set_field,
        set,
        value,
        name,
        keep_undefined=keep_undefined,
        mapper=mapper,
        camel_case_convert=camel_case_convert,
    )


def deserialize_multifield_wrapper(
    field,
    source_val,
    name,
    *,
    keep_undefined=True,
    mapper=None,
    camel_case_convert=False,
):
    """
    Only primitive values are supported, otherwise deserialization is ambiguous,
    since it can only be verified when the structure is instantiated
    """
    deserialized = source_val
    found_previous_match = False
    failures = 0
    err_messages = []
    for field_option in field.get_fields():
        try:
            ignore_none = getattr(field_option, IGNORE_NONE_VALUES, False)

            deserialized = deserialize_single_field(
                field_option,
                source_val,
                name,
                keep_undefined=keep_undefined,
                mapper=mapper,
                camel_case_convert=camel_case_convert,
                ignore_none=ignore_none,
            )
            if isinstance(field, AnyOf):
                return deserialized
            elif isinstance(field, NotField):
                raise ValueError(
                    f"{name}: Got {wrap_val(source_val)}; Matches field {field}, but must not match it"
                )
            elif isinstance(field, OneOf) and found_previous_match:
                raise ValueError(
                    f"{name}: Got {wrap_val(source_val)}; Matches more than one match"
                )
            found_previous_match = True
        except Exception as e:
            failures += 1
            err_messages.append(
                f"({len(err_messages) + 1}) Does not match {field_option}. reason: {str(e)}"
            )
            if isinstance(field, AllOf):
                raise ValueError(
                    f"{name}: Got {wrap_val(source_val)}; Does not match {field_option}. reason: {str(e)}"
                ) from e
    if failures == len(field.get_fields()) and not isinstance(field, NotField):
        raise ValueError(
            f"{name}: Got {wrap_val(source_val)}; Does not match any field option: {'. '.join(err_messages)}"
        )
    return deserialized


def deserialize_map(map_field, source_val, name, camel_case_convert=False):
    if not isinstance(source_val, dict):
        raise TypeError(f"{name}: Got {wrap_val(source_val)}; Expected a dictionary")
    if map_field.items:
        key_field, value_field = map_field.items
    else:
        key_field, value_field = None, None
    res = {}
    for key, val in source_val.items():
        ignore_none = getattr(value_field, IGNORE_NONE_VALUES, False)

        res[
            deserialize_single_field(
                key_field, key, name, camel_case_convert=camel_case_convert
            )
        ] = deserialize_single_field(
            value_field,
            val,
            name,
            camel_case_convert=camel_case_convert,
            ignore_none=ignore_none,
        )
    return res


def deserialize_single_field(  # pylint: disable=too-many-branches
    field,
    source_val,
    name="value",
    *,
    mapper=None,
    keep_undefined=True,
    camel_case_convert=False,
    ignore_none=False,
):
    """
    Deserialize a field directly, without the need to define a Structure class.
    Note the top level must be a python dict - which implies that a JSON of

    Arguments:
        field(Field):
            The field definition. For example: String, Array[Map[str, Foo]], AnyOf[Foo, Bar]
        source_val:
            the serialized value to be deserialized
        name(optional):
            name to be used for the field in case of raised exceptions
        mapper(dict): optional
            A Typedpy deserialization mapper
        keep_undefined(bool): optional
            should it create attributes for keys that don't appear in the class? default is True.

    Returns:
        a deserialized version of the data if successful, or raises an appropriate exception
    """
    if source_val is None and (ignore_none or isinstance(field, NoneField)):
        return source_val
    if isinstance(field, (Number, String, Boolean)) and not isinstance(
        field, SerializableField
    ):
        field._validate(source_val)
        value = source_val
    elif (
        isinstance(field, TypedField)
        and getattr(field, "_ty", "") in {str, int, float}
        and isinstance(source_val, getattr(field, "_ty", ""))
    ):
        value = source_val
    elif isinstance(field, Array):
        value = deserialize_array(
            field,
            source_val,
            name,
            keep_undefined=keep_undefined,
            mapper=mapper,
            camel_case_convert=camel_case_convert,
        )
    elif isinstance(field, Deque):
        value = deserialize_deque(
            field,
            source_val,
            name,
            keep_undefined=keep_undefined,
            mapper=mapper,
            camel_case_convert=camel_case_convert,
        )
    elif isinstance(field, Tuple):
        value = deserialize_tuple(
            field,
            source_val,
            name,
            keep_undefined=keep_undefined,
            mapper=mapper,
            camel_case_convert=camel_case_convert,
        )
    elif isinstance(field, Set):
        value = deserialize_set(
            field,
            source_val,
            name,
            keep_undefined=keep_undefined,
            mapper=mapper,
            camel_case_convert=camel_case_convert,
        )
    elif isinstance(field, MultiFieldWrapper):
        value = deserialize_multifield_wrapper(
            field,
            source_val,
            name,
            keep_undefined=keep_undefined,
            mapper=mapper,
            camel_case_convert=camel_case_convert,
        )
    elif isinstance(field, ClassReference):
        value = (
            deserialize_structure_internal(
                getattr(field, "_ty", None),
                source_val,
                name,
                keep_undefined=keep_undefined,
                mapper=mapper,
                camel_case_convert=camel_case_convert,
            )
            if not isinstance(source_val, Structure)
            else source_val
        )
    elif isinstance(field, StructureReference):
        try:
            value = deserialize_structure_reference(
                getattr(field, "_newclass", None),
                source_val,
                keep_undefined=keep_undefined,
                mapper=mapper,
                camel_case_convert=camel_case_convert,
            )
        except Exception as e:
            raise ValueError(f"{name}: Got {wrap_val(source_val)}; {str(e)}") from e
    elif isinstance(field, Map):
        value = deserialize_map(
            field, source_val, name, camel_case_convert=camel_case_convert
        )
    elif isinstance(field, SerializableField):
        value = field.deserialize(source_val)
    elif isinstance(field, Anything) or field is None:
        value = source_val
    elif isinstance(field, TypedField) and isinstance(source_val, (list, dict)):
        ty = getattr(field, "_ty")
        if isinstance(source_val, list):
            value = ty(*source_val)
        elif isinstance(source_val, dict):
            value = ty(**source_val)
    elif isinstance(field, NoneField):
        raise ValueError(f"{name}: Got {wrap_val(source_val)}; Expected None")
    else:
        raise NotImplementedError(
            f"{name}: Got {wrap_val(source_val)}; Cannot deserialize value of type {field.__class__.__name__}. Are "
            "you using non-Typepy class? "
        )
    return value


def deserialize_structure_reference(
    cls, the_dict: dict, *, keep_undefined, mapper, camel_case_convert=False
):
    field_by_name = {k: v for k, v in cls.__dict__.items() if isinstance(v, Field)}
    kwargs = {
        k: v for k, v in the_dict.items() if k not in field_by_name and keep_undefined
    }

    kwargs.update(
        construct_fields_map(
            field_by_name,
            keep_undefined,
            mapper,
            the_dict,
            camel_case_convert=camel_case_convert,
        )
    )
    cls(**kwargs)
    return kwargs


def construct_fields_map(
    field_by_name,
    keep_undefined,
    mapper,
    input_dict,
    camel_case_convert=False,
    ignore_none=False,
):
    result = {}
    errors = []
    mapper = mapper or {}
    for key, field in field_by_name.items():
        mapped_key = mapper.get(key, key)
        process = False
        processed_input = None
        if key in mapper:
            processed_input = get_processed_input(key, mapper, input_dict)
            if processed_input is not None:
                process = True
        elif key in input_dict and key not in mapper:
            processed_input = input_dict[key]
            process = True

        if process:
            sub_mapper = mapper.get(
                f"{mapped_key}._mapper", mapper.get(f"{key}._mapper")
            )
            if Structure.failing_fast():
                result[key] = deserialize_single_field(
                    field,
                    processed_input,
                    key,
                    mapper=sub_mapper,
                    keep_undefined=keep_undefined,
                    camel_case_convert=camel_case_convert,
                    ignore_none=ignore_none,
                )
            else:
                try:
                    result[key] = deserialize_single_field(
                        field,
                        processed_input,
                        key,
                        mapper=sub_mapper,
                        keep_undefined=keep_undefined,
                        camel_case_convert=camel_case_convert,
                        ignore_none=ignore_none,
                    )
                except (TypeError, ValueError) as ex:
                    errors.append(ex)

    raise_errs_if_needed(errors)
    return result


def deserialize_structure_internal(
    cls,
    the_dict,
    name=None,
    *,
    mapper=None,
    keep_undefined=False,
    camel_case_convert=False,
):
    """
    Deserialize a dict to a Structure instance, Jackson style.
    Note the top level must be a python dict - which implies that a JSON of
    simply a number, or string, or array, is unsupported.
    `See working examples in test. <https://github.com/loyada/typedpy/tree/master/tests/test_deserialization.py>`_

    Arguments:
        cls(type):
            The target class
        the_dict(dict):
            the source dictionary
        mapper(dict): optional
            a dict of attribute name of attribute to key in the input
        name(str): optional
            name of the structure, used only internally, when there is a
            class reference field. Users are not supposed to use this argument.
        keep_undefined(bool): optional
            should it create attributes for keys that don't appear in the class? default is False.

    Returns:
        an instance of the provided :class:`Structure` deserialized
    """

    if (
        issubclass(cls, Versioned)
        and isinstance(the_dict, dict)
        and getattr(cls, VERSION_MAPPING)
    ):
        versions_mapping = getattr(cls, VERSION_MAPPING)
        input_dict = convert_dict(the_dict, versions_mapping)
    else:
        input_dict = the_dict
    mapper = aggregate_deserialization_mappers(cls, mapper, camel_case_convert)
    if keep_undefined:
        for m in cls.get_aggregated_deserialization_mapper():
            if isinstance(m, mappers) or isinstance(mapper, mappers):
                keep_undefined = False
        if (camel_case_convert or isinstance(mapper, mappers)) and not getattr(
            cls, ADDITIONAL_PROPERTIES, False
        ):
            keep_undefined = False

    ignore_none = getattr(cls, IGNORE_NONE_VALUES, False)
    field_by_name = _get_all_fields_by_name(cls)

    if not isinstance(input_dict, dict):
        props = cls.__dict__
        fields = list(field_by_name.keys())
        required = props.get(REQUIRED_FIELDS, fields)
        additional_props = props.get(ADDITIONAL_PROPERTIES, True)
        if len(fields) == 1 and required == fields and additional_props is False:
            field_name = fields[0]
            return cls(
                deserialize_single_field(
                    getattr(cls, field_name, None),
                    input_dict,
                    field_name,
                    ignore_none=ignore_none,
                )
            )
        raise TypeError(f"{name}: Expected a dictionary; Got { wrap_val(input_dict)}")

    kwargs = {
        k: v for k, v in input_dict.items() if k not in field_by_name and keep_undefined
    }

    kwargs.update(
        construct_fields_map(
            field_by_name,
            keep_undefined,
            mapper,
            input_dict,
            camel_case_convert=camel_case_convert,
            ignore_none=ignore_none,
        )
    )

    return cls(**kwargs)


def deserialize_structure(
    cls, the_dict, *, mapper=None, keep_undefined=True, camel_case_convert=False
):
    """
    Deserialize a dict to a Structure instance, Jackson style.
    `See working examples in test. <https://github.com/loyada/typedpy/tree/master/tests/test_deserialization.py>`_

    Arguments:
        cls(type):
            The target class
        the_dict(dict):
            the source dictionary
        mapper(dict): optional
            the key is the target attribute name. The value can either be a path of the value in the source dict
            using dot notation, for example: "aaa.bbb", or a :class:`FunctionCall`. In the latter case,
            the function is the used to preprocess the input prior to deserialization/validation.
            The args attribute in the function call is optional. If non provided, the input to the function is
            the value with the same key. Otherwise it is the keys of the values in the input that are injected
            to the provided function. See working examples in the tests link above.
        keep_undefined(bool): optional
            should it create attributes for keys that don't appear in the class? default is True.

    Returns:
        an instance of the provided :class:`Structure` deserialized
    """
    return deserialize_structure_internal(
        cls,
        the_dict,
        mapper=mapper,
        keep_undefined=keep_undefined,
        camel_case_convert=camel_case_convert,
    )


SENTITNEL = uuid.uuid4()


def get_processed_input(key, mapper, the_dict):
    def _get_arg_list(key_mapper):
        vals = [deep_get(the_dict, k, default=SENTITNEL) for k in key_mapper.args]
        return [v for v in vals if v != SENTITNEL]

    key_mapper = mapper[key]
    if isinstance(key_mapper, (FunctionCall,)):
        args = _get_arg_list(key_mapper) if key_mapper.args else [the_dict.get(key)]
        processed_input = key_mapper.func(*args) if args else None
    elif isinstance(key_mapper, (str,)):
        val = deep_get(the_dict, key_mapper)
        processed_input = val if val is not None else the_dict.get(key)
    elif isinstance(key_mapper, Constant):
        processed_input = key_mapper()
    else:
        raise TypeError(
            f"mapper value must be a key in the input or a FunctionCal. Got {wrap_val(key_mapper)}"
        )
    return processed_input


# pylint: disable=too-many-return-statements
def serialize_multifield_wrapper(fields, name, val, mapper, camel_case_convert):
    for field in fields:
        try:
            if getattr(field, "_validate", None):
                field._validate(val)
            return serialize_field(field, val, camel_case_convert)
        except:  # pylint: disable=bare-except
            pass
    else:  # pylint: disable=useless-else-on-loop
        raise ValueError(f"{name}: cannot serialize value: {val}")


def serialize_val(field_definition, name, val, mapper=None, camel_case_convert=False):
    if isinstance(field_definition, SerializableField):
        return field_definition.serialize(val)
    if isinstance(field_definition, MultiFieldWrapper):
        return serialize_multifield_wrapper(
            field_definition.get_fields(), name, val, mapper, camel_case_convert
        )
    if isinstance(field_definition, (Number, Boolean, String)) or val is None:
        return val
    if isinstance(field_definition, Anything) and (
        isinstance(val, (int, float, str, bool)) or val is None
    ):
        return val
    if isinstance(field_definition, SizedCollection):
        if isinstance(field_definition, Map):
            if (
                isinstance(field_definition.items, list)
                and len(field_definition.items) == 2
            ):
                key_type, value_type = field_definition.items
                return {
                    serialize_val(
                        key_type, name, k, camel_case_convert=camel_case_convert
                    ): serialize_val(
                        value_type, name, v, camel_case_convert=camel_case_convert
                    )
                    for (k, v) in val.items()
                }
            else:
                return {
                    serialize_val(
                        Anything, name, k, camel_case_convert=camel_case_convert
                    ): serialize_val(
                        Anything, name, v, camel_case_convert=camel_case_convert
                    )
                    for (k, v) in val.items()
                }

        items = getattr(field_definition, "items", None)
        if isinstance(items, list):
            return [
                serialize_val(
                    items[ind],
                    name,
                    v,
                    mapper=mapper,
                    camel_case_convert=camel_case_convert,
                )
                for ind, v in enumerate(val)
            ]
        elif isinstance(items, Field):
            return [
                serialize_val(
                    items,
                    name,
                    i,
                    mapper=mapper,
                    camel_case_convert=camel_case_convert,
                )
                for i in val
            ]
        else:
            return [
                serialize_val(
                    None, name, i, mapper=mapper, camel_case_convert=camel_case_convert
                )
                for i in val
            ]
    if isinstance(val, (list, set, tuple)):
        return [
            serialize_val(None, name, i, camel_case_convert=camel_case_convert)
            for i in val
        ]
    if isinstance(field_definition, Anything) and isinstance(val, Structure):
        return serialize(val, mapper=mapper, camel_case_convert=camel_case_convert)
    if isinstance(val, Structure) or isinstance(field_definition, Field):
        return serialize_internal(
            val, mapper=mapper, camel_case_convert=camel_case_convert
        )

    # nothing worked. Not a typedpy field. Last ditch effort.
    try:
        return json.loads(json.dumps(val))
    except Exception as ex:
        raise ValueError(f"{name}: cannot serialize value: {ex}") from ex


def serialize_field(field_definition: Field, value, camel_case_convert=False):
    """
    Serialize a specific :class:`Field` from a structure to a JSON-like dict.
    Example:

            .. code-block:: python

                class Foo(Structure):
                    a = String
                    i = Integer

                class Bar(Structure):
                    x = Float
                    foos = Array[Foo]

                bar = Bar(x=0.5, foos=[Foo(a='a', i=5), Foo(a='b', i=1)])
                assert serialize_field(Bar.foos, bar.foos)[0]['a'] == 'a'


    Arguments:
        field_definition(:class:`Field`):
           the field definition

        value:
             the value of the field to deserialize

    Returns:
        a serialized Python object that can be directly converted to JSON
    """
    return serialize_val(
        field_definition,
        field_definition._name,
        value,
        camel_case_convert=camel_case_convert,
    )


def _get_mapped_value(mapper, key, items):
    if key in mapper:
        key_mapper = mapper[key]
        if isinstance(key_mapper, (FunctionCall,)):
            args = (
                [deep_get(items, k) for k in key_mapper.args]
                if key_mapper.args
                else [items.get(key)]
            )
            return key_mapper.func(*args)
        elif key_mapper is DoNotSerialize:
            return key_mapper
        elif isinstance(key_mapper, Constant):
            return key_mapper()
        elif not isinstance(key_mapper, (FunctionCall, str)):
            raise TypeError("mapper must have a FunctionCall or a string")

    return None


def _convert_to_camel_case_if_required(key, camel_case_convert):
    if camel_case_convert:
        words = key.split("_")
        return words[0] + "".join(w.title() for w in words[1:])
    else:
        return key


def serialize_internal(structure, mapper=None, compact=False, camel_case_convert=False):
    cls = structure.__class__
    field_by_name = _get_all_fields_by_name(cls)
    if isinstance(structure, (Structure, ClassReference)):
        mapper = aggregate_serialization_mappers(
            structure.__class__, mapper, camel_case_convert
        )
    mapper = {} if mapper is None else mapper
    if isinstance(structure, getattr(Generator, "_ty", None)):
        raise TypeError("Generator cannot be serialized")
    items = (
        structure.items()
        if isinstance(structure, dict)
        else [(k, v) for (k, v) in structure.__dict__.items() if k != "_instantiated"]
    )
    props = structure.__class__.__dict__
    fields = list(field_by_name.keys())
    additional_props = props.get(ADDITIONAL_PROPERTIES, True)
    if (
        len(fields) == 1
        and props.get(REQUIRED_FIELDS, fields) == fields
        and additional_props is False
        and compact
    ):
        key = fields[0]
        result = serialize_val(
            field_by_name.get(key, None),
            key,
            getattr(structure, key),
            camel_case_convert=camel_case_convert,
        )
    else:
        mapper = mapper or {}
        result = {}
        items_map = dict(items)
        for key, val in items:
            if val is None:
                continue
            mapped_key = (
                mapper[key]
                if key in mapper and isinstance(mapper[key], (str,))
                else _convert_to_camel_case_if_required(key, camel_case_convert)
            )
            mapped_value = _get_mapped_value(mapper, key, items_map)
            if mapped_value is not DoNotSerialize:
                the_field_definition = (
                    Anything if mapped_value else field_by_name.get(key, None)
                )
                sub_mapper = mapper.get(f"{key}._mapper", {})
                result[mapped_key] = serialize_val(
                    the_field_definition,
                    key,
                    mapped_value or val,
                    mapper=sub_mapper,
                    camel_case_convert=camel_case_convert,
                )
    if getattr(structure, "_additional_serialization", None):
        additional_props = structure._additional_serialization()
        if not isinstance(additional_props, dict):
            raise TypeError("_additional_serialization must return a dict")
        for key, value in additional_props.items():
            result[key] = value() if callable(value) else value
    return result


def serialize(value, *, mapper: Dict = None, compact=False, camel_case_convert=False):
    """
    Serialize an instance of :class:`Structure` to a JSON-like dict.

    Arguments:
        value(:class:`Structure` or a field value with an obvious serialization):
            The value to be serialized - a structure instance, or a field value for which typedpy can deduce the
            serialization.
            In the general case, if you just need to serialize a field value, it's better to use serialize_field().
        mapper(dict): optional
             a dictionary where the key is the name of the attribute in the structure, and the value is name of the
             key to map its value to, or a :class:`FunctionCall` where the function is the transformation, and
             the args are a list of attributes that are arguments to the function. if args is empty it
             function transform the current attribute.
        compact(bool):
             whether to use a compact form for Structure that is a simple wrapper of a field.
             for example: if a Structure has only one field of an int, if compact is True
             it will serialize the structure as an int instead of a dictionary

    Returns:
        a serialized Python object that can be directly converted to JSON
        :param compact: in case there is a single attribute, it does not wrap it with a dictionary
        :param structure: an instance of :class:`Structure`
        :param mapper: a dict with the new key, by the attribute name

    """
    if not isinstance(value, (Structure, StructureReference)):
        if value is None or isinstance(value, (int, str, bool, float)):
            return value
        if isinstance(value, (_ListStruct, _DictStruct)):
            field_definition = value._field_definition
            return serialize_val(
                field_definition,
                field_definition._name,
                value,
                camel_case_convert=camel_case_convert,
            )
        if isinstance(value, (enum.Enum,)):
            return value.name
        raise TypeError(
            f"serialize: Not a Structure or Field that with an obvious serialization. Got: {value}."
            " Maybe try serialize_field() instead?"
        )
    return serialize_internal(
        value, mapper=mapper, compact=compact, camel_case_convert=camel_case_convert
    )


class HasTypes:
    """
    A mixin that can be added to a base-class :class:`Structure`. It adds to the serialization of
    any instance of a subclass, its type.
    Since version 2.12.1.
    """

    def _additional_serialization(self) -> dict:
        return {"type": self.__class__.__name__.lower()}



import sys
from pathlib import Path

from typedpy.type_helpers import create_stub_for_file


def main():
    if len(sys.argv)>4 or len(sys.argv)<3:
        print(f"Usage: {sys.argv[0]} <src-root-abs-dir> <src-script-abs-path> [stubs-dir]")
        sys.exit(1)
    src_root_abs_path = sys.argv[1]
    input_file_abs_path = sys.argv[2]
    stub_dir = sys.argv[3] if len(sys.argv)==4 else ".stubs"
    src_dir_abs_path = str(Path(src_root_abs_path) / Path(stub_dir))
    create_stub_for_file(
        input_file_abs_path, src_root_abs_path, src_dir_abs_path
    )


if __name__ == "__main__":
    main()
